"""
Module for parsing docstrings and making the .json file for swagger-ui
"""

from os import path as osp
import json
import glob
import re
import yaml

from pyramid.settings import asbool

from spynl.main.version import __version__ as spynl_version
from spynl.main.utils import get_logger, get_yaml_from_docstring
from spynl.main.docs.settings import get_ini_doc_setting


extended_description = '''
All endpoints usually return application/json, unless otherwise
specified here or requested differently by the request.
They will have a "status" (ok|error) field and all error responses
also will have a "message" field.
'''


swagger_doc = {
    'swagger': '2.0',
    'info': {
        'version': spynl_version,
        'title': 'Spynl Endpoints',
        'description': 'A list of all endpoints on this Spynl instance,'
                       ' with a short description on how to use them.'
                       ' <br/><br/><span style="color: grey;">{}<span>'
                       .format(extended_description)
    },
    'paths': {},
    'definitions': {}
}


# list of endpoints where we hide the sandbox
HIDE_TRYITOUT_IDS = []


def document_endpoint(config, function, endpoint_name, resource=None):
    """parse docstring and store in swagger_doc dict"""
    log = get_logger('Spynl Documentation')

    path = '/{}'.format(endpoint_name)
    yaml_str = get_yaml_from_docstring(function.__doc__, load_yaml=False)

    if not yaml_str:
        log.warning("No YAML found in docstring of endpoint %s"
                    " (resource: %s). Cannot generate entry in /about/doc."
                    % (endpoint_name, resource))
    else:
        if resource:
            # Replace $resource in the docstring with the actual resource
            yaml_str = re.sub(r'\$resource', resource, yaml_str)
        try:
            yaml_str = insert_ini_settings(config, yaml_str)
            yaml_doc = yaml.load(yaml_str)
            # If a path has a different view for get and post, add both
            if path in swagger_doc['paths']:
                swagger_doc['paths'][path].update(yaml_doc)
            else:
                swagger_doc['paths'][path] = yaml_doc
        except yaml.YAMLError as e:
            log.error('Wrong yaml code for endpoint %s:', path)
            log.error(e)
            return

        if 'validations' in yaml_doc:
            for method in [m for m in ('get', 'post') if m in yaml_doc]:
                valdoc = "#### Validations\n"
                valdoc += "In | Schema | Apply to | Repeat\n"
                valdoc += "------ | ------ | ------ | -----\n"
                if isinstance(yaml_doc['validations'], list):
                    for val in yaml_doc['validations']:
                        if isinstance(val, dict):
                            valdoc += '{}|<a href="{}">{}</a>|{}|{}\n'\
                                    .format(val.get('in'),
                                            '/about/schemas?schema={}'
                                            .format(val.get('schema')),
                                            val.get('schema'),
                                            val.get('apply-to', '*'),
                                            val.get('repeat', 'no'))
                yaml_doc[method]['description'] += valdoc

        if 'show-try' in yaml_doc:
            if not asbool(yaml_doc['show-try']):
                for method in ('get', 'post'):
                    if method in yaml_doc:
                        for tag in yaml_doc.get(method).get('tags', []):
                            htio_id = '#{tag}_{method}{path}'\
                                    .format(tag=tag,
                                            method=method,
                                            path=path)
                HIDE_TRYITOUT_IDS.append(htio_id.replace('-', '_')
                                                .replace(' ', '_')
                                                .replace('/', '_'))

        # the 1st line of a docstring can be used for the swagger summary
        docstring = function.__doc__
        doc_lines = docstring.split("\n")
        if len(doc_lines) > 0:
            first_doc_line = doc_lines[1].strip()
            if first_doc_line == '---':
                first_doc_line = ''
        for method in [m for m in ('get', 'post')
                       if m in swagger_doc['paths'][path]]:
            path_doc = swagger_doc['paths'][path][method]
            if 'summary' not in path_doc or not path_doc['summary']:
                if resource:
                    path_doc['summary'] = re.sub(r'\$resource', resource,
                                                 first_doc_line)
                else:
                    path_doc['summary'] = first_doc_line


def make_docs(config):
    """Write swagger file. Get definitions from files first."""
    log = get_logger('Spynl Documentation')

    # add definitions from files
    def_folder = 'spynl/main/docs/definitions'
    definitions = glob.glob('%s/*.yml' % def_folder)
    for definition in definitions:
        try:
            with open(definition, 'r') as current_file:
                # check if the folder contains the definition
                match = re.match(r'%s/(.*)\.yml' % def_folder, definition)
                definition = match.group(1)
                # Read the file first, this will give a more descriptive
                # error message in case there is a yaml error.
                yaml_text = current_file.read()
                yaml_text = insert_ini_settings(config, yaml_text)
                swagger_doc['definitions'][definition] = yaml.load(yaml_text)
        except yaml.YAMLError as e:
            log.error('File %s has yaml errors:', current_file.name)
            log.error(e)
        except IOError as e:
            log.error('I/O error(%s: %s', e.errno, e.strerror)

    # turn swagger_doc into a JSON file
    swagger_file = '{}/swagger-ui/spynl.json'\
                   .format('/'.join(osp.abspath(__file__).split('/')[:-1]))
    try:
        with open(swagger_file, 'w') as outfile:
            json.dump(swagger_doc, outfile, indent=4, separators=(',', ': '),
                      sort_keys=True)
    except IOError as e:
        log.error('I/O error(%s: %s', e.errno, e.strerror)
    except (TypeError, OverflowError, ValueError) as e:
        log.error('Swagger file could not be dumped:')
        log.error(e)


def add_swagger_reusables(config, *args):
    """
    Add definitions from docstrings other than the endpoints.
    This function should be called from the plugger.py of the module.
    Arguments can be any object with a docstring, most probably a function.
    """
    log = get_logger('Spynl Documentation')

    for arg in args:
        docstring = arg.__doc__
        yaml_sep = docstring.find('---')
        tag = re.search('--(parameter|response)?:?([a-zA-Z_-]*?)--', docstring)
        yaml_description = None
        if tag and yaml_sep != -1:
            swagger_type = tag.group(1)
            name = tag.group(2)
            try:
                description = insert_ini_settings(config, docstring[yaml_sep:])
                yaml_description = yaml.load(description)
            except yaml.YAMLError as e:
                log.error('Docstring %s has yaml errors:', arg)
                log.error(e)
            if not yaml_description:
                log.error('No description could be parsed for %s:%s',
                          swagger_type, name)
        else:
            log.error('Docstring %s is missing yaml separation (---)'
                      ' or name tag (--tag--, allowed characters:'
                      ' a-z, A-Z, - and/or _)', arg)
        if swagger_type == 'parameter':
            if 'parameters' not in swagger_doc:
                swagger_doc['parameters'] = {}
            swagger_doc['parameters'][name] = yaml_description
        elif swagger_type == 'response':
            if 'responses' not in swagger_doc:
                swagger_doc['responses'] = {}
            swagger_doc['responses'][name] = yaml_description
        else:
            swagger_doc['definitions'][name] = yaml_description


def insert_ini_settings(config, description):
    """
    Fill in the value of a setting that was set in the ini file.

    The text $setting[setting_name] will be replaced by:
    actual_setting (the setting_name for this Spynl instance)
    """
    log = get_logger('Spynl Documentation')

    settings = config.registry.settings
    flag = re.compile(r'\$setting\[(?P<setting_name>[\S]*?)\]')
    match = flag.search(description)

    while match is not None:
        setting_name = match.group('setting_name')
        setting = settings.get(setting_name)
        if setting is not None:
            description = re.sub(r'\$setting\[' + setting_name + r'\]',
                                 str(setting) + r' (the ' + setting_name +
                                 ' for this Spynl instance)', description)
        else:
            default = None
            setting_doc = get_ini_doc_setting(setting_name)
            if setting_doc is not None:
                default = get_ini_doc_setting(setting_name).get('default')
            if default is not None:
                description = re.sub(r'\$setting\[' + setting_name + r'\]',
                                     str(default) + r' (the ' + setting_name +
                                     ' for this Spynl instance)', description)
            else:
                description = re.sub(r'\$setting\[' + setting_name + r'\]',
                                     setting_name, description)
                log.error('There is no value for setting %s', setting_name)
        match = flag.search(description)

    return description



"""Tasks for operations around building and deploying Spynl"""

import os
import json
from datetime import datetime, timedelta
import pytz
from urllib.parse import urlparse

import requests
from invoke import task
from invoke.exceptions import Exit

from spynl.main.pkg_utils import get_spynl_package, get_dev_config
from spynl.cli.utils import (resolve_package_names, get_spynl_package,
                             package_dir, chdir, assert_response_code) 
from spynl.cli.dev import tasks as dev_tasks


@task(help={'packages': dev_tasks.packages_help,
            'revision': 'The code revision (e.g. branch name or tag, '
                        'applicable across repositories). '
                        'Defaults to "default".',
            'fallbackrevision': 'Revision to fall back on if revision '
                                'is not given. Defaults to "default".',
            'spynlrevision': 'You can specify a revision for the spynl '
                             'repo that is used by Jenkins as basis for '
                             'the build. Defaults to "default".',
            'task': 'The AWS ECS task identifier to be restarted '
                    '(see the spynl.ops.ecr.dev_tasks setting). '
                    'Can be more than task (provide a comma-separated '
                    'list in that case). Alternatively, use "production" '
                    'to deploy to the production ECR.'})
def start_build(ctx, packages='_all', revision='default',
                fallbackrevision='default', spynlrevision='default',
                task=''):
    """Make a new Spynl build: call the SPYNL job on Jenkins."""
    config = get_dev_config()
    jenkins_url = config.get('spynl.ops.jenkins_url', '').strip()
    if not jenkins_url:
        raise Exit('[spynl ops.start_build] No spynl.ops.jenkins_url '
                   'setting found! Exiting ...')
    if os.environ.get('JENKINS_USER') is None:
        jenkins_user = os.environ.get('USER')
        print('[spynl ops.start_build] Assuming jenkins user %s ...'
              'Set env variable JENKINS_USER to change this.' % jenkins_user)
    else:
        jenkins_user = os.environ.get('JENKINS_USER')
    if os.environ.get('JENKINS_PASSWORD') is None:
        raise Exit('[spynl ops.start_build] No jenkins password given -'
                   ' please set env variable JENKINS_PASSWORD. Exiting...')

    jurl_parts = urlparse(jenkins_url)
    url = "{jscheme}://{juser}:{jpw}@{jloc}/job/Spynl-git"\
          "/buildWithParameters?delay=0sec"\
          .format(jscheme=jurl_parts.scheme, juser=jenkins_user,
                  jpw=os.environ['JENKINS_PASSWORD'], jloc=jurl_parts.netloc)
    # TODO: here we need to resolve the URLs, not the package names
    installed_packages = resolve_package_names(packages, complain_not_installed=False)
    params = dict(scm_urls=','.join(installed_packages),
                  revision=revision, fallbackrevision=fallbackrevision,
                  spynlrevision=spynlrevision, task=task)
    print('[spynl ops.start_build] Calling up Jenkins server at %s with params %s.' % (url, str(params)))
    response = requests.post(url, params)
    print('[spynl ops.start_build] Got response: %s' % response)


@task(help={'packages': dev_tasks.packages_help,
            'revision': 'The code revision (e.g. branch name or tag, '
                        'applicable across repositories). '
                        'Defaults to "default".',
            'fallbackrevision': 'Which revision to fallback on if the '
                                'repo does not have revision. '
                                'Defaults to "default".',
            'revertchanges': 'Revert uncommited changes in order to update.'
                             'Otherwise this task will raise if a repo has '
                             'local changes.'})
def mk_repo_state(ctx, packages='_all', revision='default',
                      fallbackrevision='default',
                      revertchanges=False):
    """
    Go through all repos (by package) and collect the commit IDs they are at
    in the current revision (e.g. a branch or tag).
    Store that information in a file.

    TODO: maybe we can assume we already have the package and have it installed
          at the right revision! (by dev.install)
    """
    if revision is None or revision == "":
        raise Exit("[spynl ops.mk_repo_states] revision argument is empty!")

    with open('repo-state.txt', 'w') as repo_state:
        # --- install/check for changes
        for name in resolve_package_names(packages, complain_not_installed=False):
            package = get_spynl_package(name)
            if package is None or not os.path.exists(package.location):
                # TODO: do we need this here? We should rely on prepare-stage.sh.
                #dev_tasks.install(ctx, repos=name, revision=revision,
                #                  fallbackrevision=fallbackrevision)
                pass
            with package_dir(name):
                if revertchanges:
                    ctx.run('hg revert --all')
                elif ctx.run('hg st -amdr').stdout != "":
                    raise Exit("[spynl ops.mk_package_states] Package %s has "
                               "uncomitted changes." % name)
        # --- update to revision or fallbackrevision
        for name in resolve_package_names(packages):
            with package_dir(name):
                ctx.run('hg pull -q')
                ctx.run("if echo \"$(hg branches) $(hg tags)\" | grep -i '^{rev}[[:space:]]' > /dev/null; then hg update -q {rev}; else hg update -q {fb_rev}; fi".format(rev=revision, fb_rev=fallbackrevision))
                commit_id = ctx.run('hg id -i', hide='both').stdout.strip()
                print("[spynl ops.mk_repo_states] "
                      "Package %s (at branch %s) has commit ID: %s" %
                      (name, ctx.run('hg branch', hide='both').stdout.strip(),
                       commit_id))
                repo_state.write("%s %s\n" % (name, commit_id))


@task(help={'buildnr': 'Number of the SPYNL-Deploy build of Jenkins.',
            'task': 'Can be either "production" or one of the dev '
                    'tasks given by the spynl.ops.ecr.dev_tasks setting. '
                    'With the former, the image will be pushed to the '
                    'production AWS ECR. The build number will be attached '
                    'to the tag. With the latter, the image will be pushed '
                    'to the development AWS ECR. It will be tagged with the '
                    'task name and the task in the AWS ECS will be restarted, '
                    'so the new image will be live.',
            'revision': 'The code revision (e.g. branch name or tag, '
                        'applicable across repositories). Optional - '
                        'should be left out if a repostate.txt file has '
                        'already been made.',
            'fallbackrevision': 'A fallback revision can also be given for the '
                                'case that a repo does not have revision.'})
def deploy(ctx, buildnr=None, task=None, revision=None,
           fallbackrevision='default'):
    """Build a Spynl Docker image and deploy it."""
    # --- make sure we have repostate.txt
    if not os.path.exists('repo-state.txt'):
        if revision is None:
            raise Exit("[spynl ops.deploy] No repo-state.txt found and no "
                       "revision given.")
        mk_repo_states(ctx, revision=revision,
                       fallbackrevision=fallbackrevision)
    # --- put repostates.txt into the docker build directory
    spynl = get_spynl_package('spynl')
    ctx.run('mv repo-state.txt %s/spynl/cli/ops/docker' % spynl.location)
    # --- put production.ini into the docker build directory
    config_package = get_spynl_package(config_repo)
    ctx.run('cp %s/production.ini %s/spynl/cli/ops/docker'
            % (config_package.location, spynl.location))

    # check ECR configuration
    config = get_dev_config()
    dev_ecr_profile, dev_ecr_uri =\
        config.get('spynl.ops.ecr.dev_url', '').split('@')
    if task != 'production' and not dev_ecr_uri:
        raise Exit('[spynl ops.deploy] ECR for development is not configured. '
                   'Exiting ...')
    prod_ecr_profile, prod_ecr_uri =\
        config.get('spynl.ops.ecr.prod_url', '').split('@')
    if task == 'production' and not prod_ecr_uri:
        raise Exit('[spynl ops.deploy] ECR for production is not configured. '
                   'Exiting ...')
    dev_domain = config.get('spynl.ops.dev_domain', '')

    # --- build Docker image
    with chdir(spynl.location + '/cli/ops/docker'):
        result = ctx.run('./build-image.sh %s %s' % (buildnr, dev_domain))
        if not result:
            raise Exit("[spynl ops.deploy] Building docker image failed: %s"
                       % result.stderr)
        # report the Spynl version from the code we just built
        spynl_version = ctx.run('cat built.spynl.version').stdout.strip()
        ctx.run('rm -f built.spynl.version')  # clean up
        print('[spynl ops.deploy] Built Spynl version {}'.format(spynl_version))

    if task == 'production':
        get_login_cmd = ctx.run('aws ecr --profile %s get-login '
                                '--region eu-west-1' % prod_ecr_profile)
        ctx.run(get_login_cmd.stdout.strip())
        # tag & push image with spynl_version & buildnr as tag
        ctx.run('docker tag spynl:v{v} {aws_uri}/spynl:v{v}_b{bnr}'
                .format(v=spynl_version, aws_uri=prod_ecr_uri, bnr=buildnr))
        ctx.run('docker push {aws_uri}/spynl:v{v}_b{bnr}'
                .format(v=spynl_version, aws_uri=prod_ecr_uri, bnr=buildnr))
    else:
        get_login_cmd = ctx.run('aws ecr --profile %s get-login '
                                '--region eu-west-1' % dev_ecr_profile)
        ctx.run(get_login_cmd.stdout.strip())

        ctx.run('docker tag spynl:v{v} {aws_uri}/spynl:v{v}'
                .format(v=spynl_version, aws_uri=dev_ecr_uri))
        ctx.run('docker push {aws_uri}/spynl:v{v}'
                .format(v=spynl_version, aws_uri=dev_ecr_uri))

        # tag the image for being used in one of our defined Tasks
        if task:
            ecr_dev_tasks = [t.strip() for t in config
                             .get('spynl.ops.ecr.dev_tasks', '')
                             .split(',')]
            for t in task.split(","):
                t = t.strip()
                if t not in ecr_dev_tasks:
                    raise Exit("Task: %s not found in %s. Aborting ..."
                               % (t, ecr_dev_tasks))
                print("[spynl ops.deploy] Deploying the new image for task "
                      "%s ..." % t)
                ctx.run('docker tag spynl:v{v} {aws_uri}/spynl:{task}'
                        .format(v=spynl_version, aws_uri=dev_ecr_uri, task=t))
                ctx.run('docker push {aws_uri}/spynl:{task}'
                        .format(aws_uri=dev_ecr_uri, task=t))

                # stop the task (so ECS restarts it and grabs new image)
                tasks = ctx.run('aws ecs list-tasks --cluster spynl '
                                '--service-name spynl-%s' % t)
                for task_id in json.loads(tasks.stdout)['taskArns']:
                    ctx.run('aws ecs stop-task --cluster spynl --task %s'
                            % task_id)
    ctx.run('docker logout')


@task(help={'packages': dev_tasks.packages_help})
def prepare_docker_run(ctx, packages='_all'):
    """
    Give repos a chance to make configuration changes just when the
    Docker container starts up, e.g. update /production.ini based
    on environment variables like SPYNL_ENVIRONMENT.
    This allows test/production environments to be configured in their
    preferred way.
    If the packages provide a script called "prepare-docker-run.sh",
    this task will run it.
    """
    print("[spynl ops.prepare_docker_run] Will check packages %s ..." % packages)
    for package_name in resolve_package_names(packages):
        package = get_spynl_package(package_name)
        print("[spynl ops.prepare_docker_run] Checking package %s at %s ..."
              % (package_name, package.location))
        with package_dir(package_name):
            if os.path.exists('%s/prepare-docker-run.sh' % package.location):
                print("[spynl ops.prepare_docker_run] Preparing repo: %s ..."
                      % package_name)
                ctx.run('%s/prepare-docker-run.sh' % package.location)


@task(help={'packages': dev_tasks.packages_help,
            'task': 'Spynl task, used to find out at which URL to test Spynl. '
                    ' One of the tasks specified by the ini-setting '
                    ' "spynl.ops.dev_url". For tasks other than "dev", the '
                    ' term "spynl" in the dev_url will be replaced by '
                    ' "spynl-<task>". If no task parameter is given, '
                    ' then the <url> parameter should be given.',
            'url': 'URL to run smoke test against. Will overwrite whatever '
                   'URL the task parameter indicates.'})
def smoke_test(ctx, packages="_all", url=None, task='dev'):
    """Run smoke tests"""
    config = get_dev_config()
    ecr_dev_tasks = [t.strip() for t in config
                     .get('spynl.ops.ecr.dev_tasks', '')
                     .split(',')]
    if url is not None:
        spynl_url = url.strip()
    elif task is None:
        raise Exit('Please specify a url or a task from [%s].'
                   % ecr_dev_tasks)
    else:
        if ',' in task:  # let's only do the first one
            task = task.split(',')[0].strip()
        if task not in ecr_dev_tasks:
            raise Exit('Task %s is not named in spynl.ops.ecr.dev.tasks.')
        spynl_url = config.get('spynl.ops.dev_url', '').strip()
        if not spynl_url:
            raise Exit('spynl.ops.dev_url not set.')
        if task != 'dev':
            spynl_url = spynl_url.replace('spynl', 'spynl-%s' % task)
    print("[spynl ops.smoke_test] Accessing Spynl at: [{}]".format(spynl_url))

    # Check that Spynl answers at all
    print("[spynl_dev check] pinging ...")
    res = requests.get(spynl_url + '/ping')
    assert_response_code(res, 200)
    assert res.json()['greeting'] == 'pong'

    # Check that it serves the latest build
    print("[spynl ops.smoke_test] check if build is new ...")
    res = requests.get(spynl_url + '/about/build')
    assert_response_code(res, 200)
    build_time = datetime.strptime(res.json()['build_time'],
                                   '%Y-%m-%dT%H:%M:%S%z')
    now = datetime.now(tz=pytz.UTC)
    if not now - timedelta(minutes=15) < build_time:
        raise Exit("[spynl ops.smoke_test] Build seems to be old: "
                   "%s (now - 15m) is not before %s"
                   % (now - timedelta(minutes=15), build_time))

    for package_name in resolve_packages_names(packages):
        with package_dir(package_name):
            package = get_spynl_package(package_name) 
            if os.path.exists('%s/smoke-test.py' % package.location):
                print("[spynl ops.smoke_test] Running %s/smoke-test.py ..."
                      % package.location)
                ctx.run('python %s/smoke-test.py --spynl-url %s'
                        % (package.location, spynl_url))

    print('[spynl ops.smoke_test] Spynl presence checked successfully!')



"""
Error views for 4xx and 5xx HTTP errors
"""

from pyramid.security import ACLDenied
from pyramid.httpexceptions import (HTTPForbidden, HTTPNotFound,
                                    HTTPInternalServerError)

from spynl.main.exceptions import SpynlException
from spynl.main.utils import log_error
from spynl.main.locale import SpynlTranslationString as _


def spynl_error(exc, request):
    """
    Handle raised SpynlExceptions.

    Get meta info from the assorted HTTP Error, log information and return
    a typical Spynl response.
    """
    http_exc = exc.http_escalate_as()
    request.response.status = http_exc.status
    request.response.status_int = http_exc.status_int
    request.response.content_type = 'application/json'  # this is Spynl default

    top_msg = "Spynl Error of type %s with message: '%s'."
    message = exc.message
    log_error(exc, request, top_msg, exc.__class__.__name__, message)
    return {'status': 'error',
            'type': exc.__class__.__name__,
            'message': message}


def error400(exc, request):
    """
    Handle all HTTPErrors.

    We collect information about the original error as best as possible.
    We log informaton and return a typical Spynl response.
    """
    # Set response meta data
    request.response.status = exc.status
    request.response.status_int = exc.status_int
    request.response.content_type = 'application/json'  # this is Spynl default

    error_type = exc.__class__.__name__
    if isinstance(exc, HTTPNotFound):
        message = _('no-endpoint-for-path',
                    default="No endpoint found for path '${path}'.",
                    mapping={'path': request.path_info})
    elif (isinstance(exc, HTTPForbidden) and hasattr(exc, 'result') and
          exc.result is not None):
        if isinstance(exc.result, ACLDenied):
            message = _(
                'permission-denial',
                default="Permission to '${permission}' ${context} was denied.",
                mapping={'context':request.context.__class__.__name__,
                         'permission': exc.result.permission})
            emeta = exc.result  # TODO: log this as detail info
        else:
            message = exc.result.msg
    elif isinstance(exc, HTTPInternalServerError):
        message = _('internal-server-error',
                    default='An internal server error occured.')
    else:
        message = exc.explanation
        if exc.detail:
            if ":" in exc.detail:
                error_type, message = exc.detail.split(':', 1)
            else:
                message = exc.detail

    if error_type.endswith('Exception'):
        error_type = error_type[:-len('Exception')]

    top_msg = "HTTP Error of type %s with message: '%s'."

    log_error(exc, request, top_msg, error_type, message)

    response = {'status': 'error', 'type': error_type, 'message': message}
    if hasattr(exc, 'details') and exc.details:
        response['details'] = exc.details

    return response


def error500(exc, request):
    """
    Handle all failures we do not anticipate in error.

    Give back json, set the error status to 500,
    and only include minimal information (to decrease attack vector).
    However, we log all information we can about the error for debugging.
    """
    # First set some response metadata
    request.response.status_int = 500
    request.response.content_type = 'application/json'  # this is Spynl default

    message = str(exc)
    if not message:
        message = _('no-message-available', default="No message available.")

    error_type = exc.__class__.__name__
    if error_type.endswith('Exception'):
        error_type = error_type[:-len('Exception')]

    top_msg = "Server Error (500) of type '%s' with message: '%s'."

    log_error(exc, request, top_msg, error_type, message)

    message = _('internal-server-error',
                default='An internal server error occured.')
    return dict(status='error', message=message)



"""
This module provides information for version, db, build and enviroment.
"""

import sys
import os
from subprocess import check_output
import pip

from pyramid.renderers import render
from pyramid.httpexceptions import HTTPFound

from spynl.main.version import __version__ as spynl_version
from spynl.main.utils import get_settings
from spynl.main.locale import SpynlTranslationString as _
from spynl.main.dateutils import now, date_to_str
from spynl.main.docs.settings import ini_doc, ini_description
from spynl.main.docs.documentation import HIDE_TRYITOUT_IDS
from spynl.main.pkg_utils import (get_spynl_package,
                                  get_dev_config,
                                  get_spynl_packages)


def hello(request):
    """
    The index for all about-endpoints.

    ---
    get:
      description: >
        The index for all about-endpoints.

        ####Response

        JSON keys | Content Type | Description\n
        --------- | ------------ | -----------\n
        status    | string | 'ok' or 'error'\n
        message   | string | Information about the available about/*
        endpoints.\n
        spynl_version | string | The version of this Spynl instance.\n
        language  | string | The language (e.g. "en") served.\n
        time      | string | Time in format: $setting[spynl.date_format]\n

      tags:
        - about
      show-try: true
    """
    return {'message': _('about-message',
                         default='This is the Spynl API/Middleware. '
                                 'You can get more information at '
                                 'about/endpoints, about/ini, about/versions, '
                                 'about/build and about/environment.'),
            'spynl_version': spynl_version,
            'language': request._LOCALE_,
            'time': date_to_str(now())}


def endpoint_doc(request):
    """
    All endpoints offered by this instance, with explanations.

    ---
    get:
      tags:
        - about
      description:
        Presented as text/html (in swagger-ui).
        Requires 'read' permission for the 'about' resource.
    show-try: No
    """
    request.response.content_type = 'text/html'
    path2swagger = '{}/../docs/swagger-ui/index.html'\
                   .format('/'.join(os.path.abspath(__file__).split('/')[:-1]))
    index = open(path2swagger, 'r').read()
    static_url = request.static_url('spynl.main:docs/swagger-ui/')
    # We would like to treat the swagger-ui directory as a drop-in
    # replacable dir, so we simply re-create the HTML by doing some
    # replacements.
    index = index.replace("<title>Swagger UI",
                          "<title>Spynl Endpoints Documentation")
    index = index.replace("http://petstore.swagger.io/v2/swagger.json",
                          "{}spynl.json".format(static_url))
    index = index.replace("href='css", "href='{}css".format(static_url))
    index = index.replace("src='", "src='{}".format(static_url))
    # We'll now add some CSS before the body closes
    # We hide the Swagger head
    css = '#header {display:none; !important}'
    # We hide the parameter sandbox for some views
    for htio_id in HIDE_TRYITOUT_IDS:
        css += '{} .sandbox {{display:none; !important}}'.format(htio_id)
    ins = index.find('</body>')
    index = index[:ins] + '<style>' + css + '</style></body></html>'

    return index


def schemas(request):
    """
    Using docson to display any relevant JSON schemas used by Spynl.

    ---
    get:
      tags:
        - about
      description: >
        Schema files are read from an internal Spynl location and
        Presented as text/html (using docson).
        This endpoint requires 'read' permission for the 'about' resource.
        It is linked from by /about/doc when schemas are used for
        validations by an endpoint.

        ####Parameters

        JSON keys |   Type   |    Req.     | Description\n
        --------- | -------- | ----------- | --------- \n
        schema    |   string |  &#10004;   | Name of the schema file. \n

    show-try: No
    """
    request.response.content_type = 'text/html'
    static_url = request.static_url('spynl.main:docs/docson/')
    schema = request.args.get('schema')
    if not schema.endswith('.json'):
        schema = schema + '.json'
    return HTTPFound("{}#schemas/{}".format(static_url, schema))


def versions(request):
    """
    The changeset IDs of Spynl and all installed plugins.

    ---
    get:
      tags:
        - about
      description: >
        Requires 'read' permission for the 'about' resource.

        ####Response

        JSON keys | Content Type | Description\n
        --------- | ------------ | -----------\n
        status    | string | 'ok' or 'error'\n
        plugins   | dict   | "spynl.plugin": "SCM checkin id" for each Spynl
        plugin. Currently works only with mercurial.\n
        spynl_version | string | The version of this Spynl instance.\n
        time      | string | time in format: $setting[spynl.date_format]\n

    """
    spynl_settings = get_settings()
    dev_config = get_dev_config()

    response = dict(spynl_version=spynl_version)
    response['time'] = date_to_str(now())

    # Plugins
    response['plugins'] = {}
    packages = get_spynl_packages()
    cmd = 'git log --format="%H" -n 1'
    if dev_config.get('scm_type') == 'hg':
        cmd = 'hg id -i 2>&1'
    # TODO: if pip-installed, we'll have to ask pip explicitly for spynl
    spynl = get_spynl_package('spynl', packages)
    try:
        os.chdir(spynl.location)
        response['plugins']['spynl'] = \
                check_output(cmd, shell=True).strip()
    except Exception as err:
        response['plugins']['spynl'] = str(err)
    for package in packages:
        try:
            os.chdir(package.location)
            response['plugins'][package.project_name] = \
              check_output(cmd, shell=True).strip()
        except Exception as err:
            response['plugins'][package.project_name] = str(err)
    return response


def build(request):
    """
    Information about the build of this instance.

    ---
    get:
      tags:
        - about
      description: >

        ####Response

        JSON keys | Content Type | Description\n
        --------- | ------------ | -----------\n
        status    | string | 'ok' or 'error'\n
        build_time| string | time in format: $setting[spynl.date_format]\n
        start_time| string | time in format: $setting[spynl.date_format]\n
        build_number | string | The build number, set by Jenkins\n
        spynl_function| string | Which functionality this node has been spun up
        to fulfil\n
        time      | string | time in format: $setting[spynl.date_format]\n
    """
    spynl_settings = get_settings()
    response = {}

    response['time'] = date_to_str(now())
    response['build_time'] = spynl_settings.get('spynl.ops.build_time', None)
    response['start_time'] = spynl_settings.get('spynl.ops.start_time', None)
    response['build_number'] = spynl_settings.get('spynl.ops.build_number', None)
    response['spynl_function'] = spynl_settings.get('spynl.ops.function', None)

    return response


def environment(request):
    """
    All software packages installed by pip for this instance.

    ---
    get:
      tags:
        - about
      description: >
        Requires 'read' permission for the 'about' resource.

        ####Response

        JSON keys | Content Type | Description\n
        --------- | ------------ | -----------\n
        status    | string | 'ok' or 'error'\n
        time      | string | time in format: $setting[spynl.date_format]\n
        python    | string | Python version\n
        pip-installed-packages | dict | "name":"version" for each package.\n
    """
    response = dict(time=date_to_str(now()))
    vi = sys.version_info
    response['python'] = "%s.%s.%s" % (vi.major, vi.minor, vi.micro)
    response['pip-installed-packages'] = {}
    installed_packages = pip.get_installed_distributions()
    for package in [p for p in installed_packages
                    if not p.key.startswith('spynl')]:
        response['pip-installed-packages'][package.key] = package.version

    return response


def ini(request):
    """
    A table of all Spynl ini settings in this instance.

    ---
    get:
      tags:
        - about
      description: >
        Presented as a text/html table. Columns for name of the setting,
        the value in this instance, whether it is required, default value
        and text description.


        Requires 'read' permission for the 'about' resource.
    """
    spynl_settings = get_settings()
    for i, setting in enumerate(ini_doc):
        ini_doc[i]['value'] = str(spynl_settings.get(setting['name']))

    request.response.content_type = 'text/html'
    result = render('spynl.main:docs/ini.jinja2',
                    {'settings': ini_doc,
                     'description': ini_description},
                    request=request)
    return result



"""
Routing - make it possible to add endpoints and document them.
For endpoints that are based on resources, it is important
that the right path leads to the right resource being used
as context in the request.
Spynl allows each Resource being accessible with more than
paths (aliases). We use the URLDispatch feature of Pyramid
here, so we create routes for all possible cases.
"""


from collections import namedtuple

from pyramid.security import NO_PERMISSION_REQUIRED

from spynl.main.utils import get_logger, is_production_environment
from spynl.main.docs import document_endpoint


class Resource:
    """
    Spynl Resource class, which handles only a name property.

    Inherit from this class to produce a resource context
    for config.add_endpoint.
    """

    def __init__(self, request):
        """by accepting a request, this class is now a context factory"""
        pass

    paths = []
    """list here the URL paths under which the resource will be visible"""

    available_in_production = True
    """Set this to False if you want this resource to be seen in dev only"""

    @property
    def name(self):
        """
        This property is the name of the path under which Spynl uses this
        resource
        """
        return self.__class__.__name__.lower()


def add_resource(config, resource_class):
    """
    Add a resource which Spynl handles to a dict.
    Also add routes for this resource (for all its paths),
    such that the endpoints defined with it as context will
    work and that Spynl will use the resource's ACLs to
    authorise them.
    """
    logger = get_logger('spynl.main.routing')

    if not issubclass(resource_class, Resource):
        raise Exception("The resource cannot be added, as it is not a "
                        "subclass of spynl.main.routing.Resource")
    resource_name = resource_class.__name__
    resources = config.get_settings()['spynl.resources']

    added_paths = []
    if 'paths' in resource_class.__dict__:  # only add paths on this very class
        for path in resource_class.paths:
            added_paths.append(path)
            for ex_res in resources.values():  # check if other resources
                                               # claimed this path already
                if path in ex_res.paths:
                    logger.warn('Cannot add path "%s" for resource %s, as it'
                                ' was already added by resource %s!' %
                                (path, resource_name, ex_res.__name__))
                    added_paths.pop()
                    break
            if path in added_paths:
                route_info = config.get_settings()['spynl.resource_routes_info']
                for rpattern, route_info in route_info.items():
                    if issubclass(resource_class, route_info['resources']):
                        rf = route_info['route_factory']
                        rf(config,
                           route_name=rpattern.replace('{path}', path),
                           resource_class=resource_class, path=path)
                        logger.info("Adding routes for resource: '%s',"
                                    " path '%s', pattern '%s'",
                                    resource_name, path, rpattern)
    if len(added_paths) > 0:
        resources[resource_name] = resource_class


def add_resource_routes(config, route_name, resource_class, path):
    """
    Add routes to the config for a resource_class and path.
    We also add a route for no method passed.
    """
    config.add_route(name=route_name,
                     pattern='/%s/{method}' % path,
                     factory=resource_class)
    config.add_route(name=route_name + '.nomethod',
                     pattern='/%s' % path,
                     factory=resource_class)


def add_endpoint(config, func, endpoint_name,
                 permission=NO_PERMISSION_REQUIRED, context=None,
                 available_in_production=True, **kw):
    """
    Add an endpoint configuration.
    The context param is the resource class which will act as context
    factory in requests (we keep to the name "context" here, because
    it is being used as a keyword param in add_view).
    If None (not given), this view is basic, otherwise a resource-based
    endpoint. For a resource-based endpoint, the endpoint_name describes
    the method, but it can be empty or "/", meaning a default method
    is added (usually for GET and POST).
    All routes your endpoint should be applied to must be known at
    this point, so if a Spynl plugin adds a new route scheme, include
    (require) it before you add endpoints.
    """
    logger = get_logger('spynl.main.routing')
    if (is_production_environment(or_test=True)
            and not available_in_production):
        logger.warn("Skipping endpoint '%s' - should not be available in "
                    "production environments", endpoint_name)
        return
    if 'renderer' not in kw:
        kw['renderer'] = 'spynls-renderer'
    if context:  # This is a resource-based endpoint
        context_name = context.__name__
        if (is_production_environment(or_test=True)
                and not context.available_in_production):
            logger.warn("Not adding endpoint %s, as its resource (%s) is "
                        "not to be used in a production-like environment." %
                        (endpoint_name, context_name))
            return
        # add resource and routes if we haven't yet
        resources = config.get_settings()['spynl.resources']
        if not context_name in resources.keys():
            add_resource(config, context)
        # each resource should define their own paths
        if 'paths' not in context.__dict__.keys() or len(context.paths) == 0:
            logger.warn("Resource %s defines no paths, cannot add endpoint." %
                        context_name)
            return
        has_method = endpoint_name and endpoint_name != '/'
        match_param = has_method and "method={}".format(endpoint_name) or None
        for path in context.paths:
            rroutes = config.get_settings()['spynl.resource_routes_info']
            for route_name, rinfo in rroutes.items():
                if issubclass(context, rinfo['resources']):
                    route_name = route_name.replace('{path}', path)
                    if not has_method:
                        route_name = route_name + '.nomethod'
                    config.add_view(func, context=context,
                                    route_name=route_name,
                                    match_param=match_param,
                                    permission=permission, **kw)
                    logger.info("Added endpoint '%s' for route '%s', resource '%s'",
                                endpoint_name, route_name, context_name)
                    document_endpoint(config, func,
                                      has_method and path + '/' + endpoint_name\
                                          or endpoint_name,
                                      resource=path)
    else:  # This is a general endpoint without route
        logger.debug("Adding endpoint '%s'", endpoint_name)
        config.add_view(func, name=endpoint_name,
                        permission=permission,
                        **kw)
        document_endpoint(config, func, endpoint_name)



def main(config):
    """Define our list of resources and the means to add an endpoint."""
    resources = {}
    '''A dictionary of the resource classes Spynl serves with at least
       one endpoint. Keys are class names.'''
    config.add_settings({'spynl.resources': resources})
    resource_routes_info = {'spynl.{path}':
                            {'resources': (Resource,),
                             'route_factory': add_resource_routes}}
    '''This dict maps patterns of resource route names (with path
       being replacable) which are available to a factory function that
       adds these routes and a list of resource classes these routes
       should be created for. The function will receive config, route_name,
       resource_class and path parameters. This is extensible by other
       Spynl plugins. Our convention is that there is also a route
       pattern that ends with ".nomethod"'''
    config.add_settings({'spynl.resource_routes_info': resource_routes_info})

    config.add_directive('add_endpoint', add_endpoint)



"""
This allows to receive data in csv format.

Note: The .csv format is limited in structure, so the assumption is that the
data is flat! For data querying, this works well for all queries which return a
list of documents with the same structure and no nested documents.
"""

import csv
import io

from spynl.main.serial import objects
from spynl.main.serial import json as spynl_json
from spynl.main.serial.exceptions import MalformedRequestException


def loads(body, headers={}, context=None):
    """
    Parse CSV input. Header fields can contain delimiter and quotechar info.
    Search queries need to remain powerful in structure, so we test for JSON
    first (this can go later, when SWPY-295 is done).
    """
    if spynl_json.sniff(body):
        return spynl_json.loads(body, headers)

    # check x-spynl headers
    delimiter = headers.get('x-spynl-delimiter')
    quotechar = headers.get('x-spynl-quotechar')
    # if not given, sniff delimiter and quotechar
    if not delimiter or not quotechar:
        try:
            dialect = csv.Sniffer().sniff(body[:3000])
        except Exception as err:
            if str(err) == 'Could not determine delimiter':
                dialect = csv.Sniffer().sniff(body[:6000], delimiters=',\t|')
            else:
                raise MalformedRequestException('text/csv', str(err))

        if not delimiter:
            delimiter = dialect.delimiter
        if not quotechar:
            quotechar = dialect.quotechar

    data = body.split("\n")
    dict_data = [objects.SpynlDecoder(context=context)(dic)
                 for dic in csv.DictReader(data, delimiter=delimiter,
                                           quotechar=quotechar)]

    return {'data': dict_data}


class CSVString:
    """Very simple Dummy class to create a CSV String."""

    def __init__(self):
        self.data = ''

    def __str__(self):
        return self.data

    def write(self, s):
        """Only needed method for use in CSVWriter."""
        self.data += s


class UnicodeStringCSVWriter:
    """A CSV writer to write rows to a string, which is encoded in UTF-8."""

    def __init__(self, dialect=csv.excel, **kwds):
        # Redirect output to a queue
        self.output = io.StringIO()
        self.writer = csv.writer(self.output, dialect=dialect, **kwds)

    def writerow(self, row):
        """Write a row."""
        self.writer.writerow([str(s) for s in row])

    def release(self):
        """Fetch the UTF-8 output from the `output` and decode it."""
        data = self.output.getvalue()
        self.output.truncate(0)  # empty queue
        return data

    def writerows(self, rows):
        """Write rows."""
        for row in rows:
            self.writerow(row)


def dumps(body, pretty=False):  # pylint: disable=unused-argument
    """
    Dump the passed body into JSON.

    CSV is a flat format, and eveything nested will be send as a
    quoted dumped JSON string. This is also why all other strings
    will be quoted.
    All content should be contained in a dictionary,
    in a list under the single key "data".
    If there is no "data" key, we render the whole response as JSON.
    """
    if body.get('data'):
        csv_writer = UnicodeStringCSVWriter(quoting=csv.QUOTE_MINIMAL,
                                            quotechar="'")
        json_data = body.get('data')
        keys = list(json_data[0].keys())
        csv_writer.writerow(keys)
        for doc in json_data:
            row = [spynl_json.dumps(doc.get(k, ' ')) for k in keys]
            csv_writer.writerow(row)
        return csv_writer.release()
    return spynl_json.dumps(body, pretty)


def sniff(body):
    """Not implemented for csv."""
    return False



"""Custom (de)serialisation Exceptions"""


from spynl.main.exceptions import SpynlException
from spynl.main.locale import SpynlTranslationString as _


class UndeterminedContentTypeException(SpynlException):
    """Undetermined Content Type"""

    def __init__(self):
        """Exception message"""
        self.message =  _('undetermined-content-type-exception',
                          default='The request carries a body but the content type cannot be '
                          'determined.')


class UnsupportedContentTypeException(SpynlException):
    """Unsupported Content Type"""

    def __init__(self, content_type):
        """Exception message"""
        self.message =  _('unsupported-content-type-exception',
                          default='Unsupported content type: "${type}"',
                          mapping={'type': content_type})


class DeserializationUnsupportedException(SpynlException):
    """Deserialisation not supported"""

    def __init__(self, content_type):
        """Exception message"""
        self.message = _('deserialization-unsupported-exception',
                         default='Deserialization for content type "${type}" is '
                         'unsupported.',
                         mapping={'type': content_type})


class SerializationUnsupportedException(SpynlException):
    """Serialization not supported"""

    def __init__(self, content_type):
        """Exception message"""
        self.message = _('serialization-unsupported-exception',
                         default='Serialization for content type: "${type}" is '
                         'not supported.',
                         mapping={'type': content_type})
        return _('serialization-unsupported-exception',
                 default='Serialization for content type: "${type}" is '
                         'not supported.',
                 mapping={'type': self.args[0]})


class MalformedRequestException(SpynlException):
    """Malformed reqeust - first give content type, then message"""

    def __init__(self, content_type, request=None):
        """Exception message"""
        if request:
            self.message = _('malformed-request-exception-type',
                             default='Malformed "${type}" request: ${request}',
                             mapping={'type': content_type, 'request': request})
        else:
            self.message = _('malformed-request-exception',
                             default='Malformed request: ${type}',
                             mapping={'type': content_type})



"""
Translation logic.
Kept in its own module so it can be imported by any Spynl code.
"""
import inspect
from pyramid.i18n import TranslationString, make_localizer
from pyramid import threadlocal
from pyramid_jinja2.i18n import GetTextWrapper


class SpynlTranslationString(object):
    """
    A wrapper class around a Pyramid TranslationString.

    This class contains an entry for a Pyramid TranslationString
    (self.translation_string), and determines the domain for the string during
    initialization.

    Why we implement this:
    During serialisation, the json encoder recognizes a TranslationString as a
    str, and treats it accordingly. This means that it takes the immutable
    string that is made during creation of the TranslationString, which
    corresponds to the msgid. There is no way to overwrite this behaviour, by
    inheriting the TranslationString class itself.
    Because we have a class here that inherits from object, we can add a
    custom encoder for this class to supplement the JSONEncoder.
    """
    # TODO: should we add __slots__ = ('translation_string')?
    def __init__(self, msgid, default=None, mapping=None, context=None):
        """
        Initialize a TranslationString, using the correct domain.
        """
        # Determine the domain, use spynl.main if no domain can be determined:
        try:
            call_stack = inspect.stack()[1]
            calling_plugin = inspect.getmodule(call_stack[0])
            domain = '.'.join(calling_plugin.__name__.split('.')[:2])
        except Exception:
            domain = 'spynl.main'
        # loop over mapping dictionary to check for translation strings:
        # (In principle mappings shouldn't be translation strings, but this
        #  may be unavoidable when e.g. a broad exception gets a more detailed
        #  message from the code that raises it)
        if mapping:
            for key in mapping:
                if isinstance(mapping[key], SpynlTranslationString):
                    mapping[key] = mapping[key].translate()
        # Initialize a Pyramid TranslationString
        self.translation_string = TranslationString(msgid, default=default,
                                                    mapping=mapping,
                                                    context=context,
                                                    domain=domain)

    def __repr__(self):
        """
        Return 'msgid: interpolated string'
        """
        # TODO: maybe we should include the mapping and default explicitly
        # instead of in the interpolated message.
        msgid = str(self.translation_string)
        message = self.translation_string.interpolate()
        return '{}: {}'.format(msgid, message)

    def __str__(self):
        """
        Return the interpolated default string.
        """
        return self.translation_string.interpolate()

    def translate(self, localizer=None):
        """ Translate the translation string using the provided localizer. """
        # We prefer to get the localizer from the place where the string gets
        # tranlated, but in some places, that adds too much code (e.g. in
        # exceptions that need to translate a translation string before adding
        # it to another translation string in a mapping.
        if not localizer:
            request = threadlocal.get_current_request()
            if request is None:  # we cannot find out locale of the user, so
                return str(self) # we'll use the interpolated default
            else:
                localizer = request.localizer
        return localizer.translate(self.translation_string)
    

class TemplateTranslations(GetTextWrapper):

    def gettext(self, *args, **kwargs):
        """Implements jinja.ext.i18n `gettext` function."""
        tsf = TranslationString(*args, domain=self.domain, **kwargs)
        return self.localizer.translate(tsf)



"""
Custom handling of requests and responses.
"""

from os.path import basename, splitext
from copy import deepcopy

from pyramid.events import (NewRequest, BeforeTraversal, ContextFound,
                            NewResponse)
from pyramid.httpexceptions import HTTPUnsupportedMediaType

from spynl.main.utils import (unify_args, get_logger, is_origin_allowed,
                              get_settings)
from spynl.main.serial.typing import negotiate_request_content_type
from spynl.main.serial.exceptions import UnsupportedContentTypeException


def prepare_content_types(event):
    """
    Negotiate requested content type and pre-set
    response content type
    """
    request = event.request

    # negotiate request content type
    try:
        request.content_type = negotiate_request_content_type(request)
    except UnsupportedContentTypeException as e:
        request.content_type = None
        raise HTTPUnsupportedMediaType(detail=e.message.translate(
            request.localizer))

    # Overwrite the HTML assumption made by the browsers
    # about response type with the Spynl default, views can overwrite
    if request.response.content_type == 'text/html'\
      and '/static' not in request.path_url:
        request.response.content_type = 'application/json'


def corsify_response(event):
    """
    Add headers to the response, in order to allow
    Cross-Origin Resource Sharing (CORS).
    Then Spynl is usable from different domains. Check the ini settings
    spynl.tld_origin_whitelist and spynl.dev_origin_whitelist for info
    on how to whitelist domains.

    Helpful links:
    http://www.w3.org/TR/cors/
    https://developer.mozilla.org/en-US/docs/Web/HTTP/Server-Side_Access_Control
    http://www.kinvey.com/bloog/60/kinvey-adds-cross-origin-resource-sharing-cors
    """
    response = event.request.response
    origin = event.request.headers.get('Origin')
    if origin:  # otherwise we are on localhost or are called directly
        if is_origin_allowed(origin):
            response.headerlist.append(('Access-Control-Allow-Origin', origin))
        else:
            response.headerlist.append(('Access-Control-Allow-Origin', 'null'))
    response.headerlist.append(('Access-Control-Allow-Credentials', 'true'))
    response.headerlist.append(('Vary', 'Accept-Encoding, Origin'))


def split_extension(event):
    """Split extension off the path info"""
    request = event.request
    # tweak path info for routing
    path, ext = splitext(request.path_info)
    if not path.startswith('/static'):
        request.original_path_info = request.path_info
        request.path_info = path
    request.path_extension = ext


def store_accepted_lang(event):
    """
    Store locale on request for pyramid.i18n.negotiate_locale_name.
    We do not support dialects at the moment, so for instance,
    "en-gb" will be treated as "en".

    We look for information about the locale in this order:
    1. The "lang" cookie
    2. The request header Accept-language
    3. The first of the supported languages (spynl.languages)
    4. We fall back to "en", which is the language the code uses
    """
    request = event.request
    settings = get_settings()

    language = None

    def langcode(langstr):
        """return only up to the first two characters"""
        if len(langstr) < 2:
            return langstr
        return langstr[:2]

    supported_languages = [langcode(lang.strip()) for lang in
                           settings.get('spynl.languages', '').split(',')]
    preferred_language = supported_languages[0]

    if 'lang' in request.cookies:
        language = langcode(request.cookies['lang'])
    elif request.accept_language:
        language = langcode(request.accept_language\
                                   .best_match(supported_languages,
                                               preferred_language))
    if language is None or language not in supported_languages:
        language = preferred_language
    if language is None:
        language = 'en'
    # pylint: disable=W0212
    request._LOCALE_ = language


def parse_args_and_log_request(event):
    """
    Parse request data when context is known,
    and log the request (together with parsed arguments) with
    log level INFO.
    Best to do this when the context is known and available
    through the request object, as this can influence how
    we parse request data.
    """
    request = event.request
    # bring all args together in the args dict on the request
    request.args = unify_args(request)
    # log the request
    log = get_logger()
    args = deepcopy(request.args)
    if 'password' in args:
        args['password'] = '*********'
    origin = request.headers.get('Origin', '')
    origin_txt = ''
    if origin:
        origin_txt = ' Origin: {}'.format(origin)
    log.info('New request for URL path "%s" from %s',
             request.path_url, origin_txt,
             extra=dict(meta=dict(url=request.path_url), payload=args))


def enforce_response_type(event):
    """
    Support two exotic types of response data: text and download, which can
    be enforced by arguments.
    In the first case, we set the content type, in the second we set headers.
    """
    r = event.request
    args = getattr(r, 'args', {})

    if 'force_text' in args:
        event.response.content_type = 'text/plain'

    if 'force_download' in args:
        filename = str(basename(r.path_info))
        event.response.headers.update(
            {'Content-Description': 'File Transfer',
             'Content-Type': 'application/octet-stream',
             'Content-Disposition': 'attachment: filename=' + filename,
             'Content-Transfer-Encoding': 'binary',
             'Expires': '0',
             'Cache-Control': 'must-revalidate',
             'Pragma': 'public'})


def main(config):
    """Subscribe to pyramid events."""
    config.add_subscriber(split_extension, NewRequest)
    config.add_subscriber(store_accepted_lang, NewRequest)
    config.add_subscriber(prepare_content_types, BeforeTraversal)
    config.add_subscriber(parse_args_and_log_request, ContextFound)
    config.add_subscriber(enforce_response_type, NewResponse)
    config.add_subscriber(corsify_response, NewResponse)
